severity  = gentle
verbose = [%p] %m at line %l, column %c. %e (Severity: %s)\n
include   = RequirePodSections
color-severity-lowest = bright_blue
color-severity-low = bright_green
color-severity-medium = bright_yellow
color-severity-high = rgb530
color-severity-highest = bright_red

[TestingAndDebugging::RequireUseStrict]
equivalent_modules = Moo Modern::Perl

[TestingAndDebugging::RequireUseWarnings]
equivalent_modules = Moo Modern::Perl

[Documentation::RequirePodSections]
lib_sections = NAME | VERSION | SYNOPSIS | DESCRIPTION | BUGS AND LIMITATIONS | AUTHOR | COPYRIGHT AND LICENSE
script_sections = NAME | USAGE | OPTIONS | EXIT STATUS | BUGS AND LIMITATIONS | AUTHOR | COPYRIGHT AND LICENSE

[Subroutines::ProhibitExcessComplexity]
max_mccabe = 20

[Subroutines::RequireFinalReturn]
terminal_methods = throw
terminal_funcs = throw_data_wrongtype throw_getopt_error throw_cli_usage throw_save_error throw_no_save throw_no_handler throw_abstract throw_mod_load_failure throw_mod_run_failure throw_no_run throw_autoload_fail throw_nostruct throw_badstruct throw_template throw_twiki_no_root throw_twiki_not_found throw_twiki_require throw_twiki_no_config throw_twiki_config_missing throw_twiki_oops throw_field_not_specified

